%% -*- erlang-indent-level: 2 -*-
%%------------------------------------------------------------------------------

%%====================================================================
%% Tags for the kind of encoded types.
%%====================================================================

-define(atom_lit_tag, atom_lit).
-define(integer_lit_tag, integer_lit).
-define(integer_tag, integer).
-define(nil_tag, nil).
-define(any_tag, any).
-define(atom_tag, atom).
-define(float_tag, float).
-define(tuple_tag, tuple).
-define(list_tag, list).
-define(nonempty_list_tag, nonempty_list).
-define(union_tag, union).
-define(range_tag, range).
-define(bitstring_tag, bitstring).
-define(neg_inf, neg_inf).
-define(pos_inf, pos_inf).
-define(remote_tag, remote).
-define(local_tag, local).
-define(record_tag, record).
-define(function_tag, function).
